import dataclasses
import logging
import math
from dataclasses import dataclass
from typing import List, Optional

import torch

from olmo_core.exceptions import OLMoConfigurationError
from olmo_core.optim import INITIAL_LR_FIELD, LR_FIELD, SkipStepAdamW
from olmo_core.optim.scheduler import WSD, ConstantScheduler, Scheduler

from ..common import Duration
from ..train_module import TransformerPipelineTrainModule, TransformerTrainModule
from .callback import Callback
from .speed_monitor import SpeedMonitorCallback

log = logging.getLogger(__name__)


@dataclass
class BatchSizeSchedulerCallback(Callback):
    """
    A callback for setting a batch size scheduler over the course of a training run.
    Also adjusts the base learning rate with Adam optimizers for transformer train modules by a factor of
    ``sqrt(new_batch_size / current_batch_size)``.
    """

    batch_sizes: List[int] = dataclasses.field(default_factory=list)
    """
    Defines the batch sizes to apply, in order.
    """

    schedule: List[Duration] = dataclasses.field(default_factory=list)
    """
    Defines the schedule at which to apply each batch size.
    """

    enabled: bool = True

    def __post_init__(self):
        if len(self.batch_sizes) != len(self.schedule):
            raise OLMoConfigurationError(
                "batch_sizes and schedules should have the same number of items"
            )

        if not self.schedule or not self.enabled:
            return

        if len({duration.unit for duration in self.schedule}) > 1:
            raise OLMoConfigurationError(
                "batch size scheduler must use consistent units for all points in the schedule"
            )

        batch_size, event_start = self.batch_sizes[0], self.schedule[0]
        if event_start.value != 0:
            raise OLMoConfigurationError("batch size schedule must start at 0")

        for next_batch_size, next_event_start in zip(self.batch_sizes[1:], self.schedule[1:]):
            if next_event_start.value <= event_start.value:
                raise OLMoConfigurationError(
                    "subsequent events in the batch size schedule must be configured to occur after previous events"
                )

            if next_batch_size > batch_size:
                if next_batch_size % batch_size != 0:
                    raise OLMoConfigurationError(
                        "invalid batch size schedule, batch size can only be increased by an integer factor"
                    )
            elif next_batch_size < batch_size:
                if batch_size % next_batch_size != 0:
                    raise OLMoConfigurationError(
                        "invalid batch size schedule, batch size can only be decreased by an integer factor"
                    )
            else:
                raise OLMoConfigurationError(
                    "invalid batch size schedule, duplicate batch size in schedule"
                )

            batch_size = next_batch_size
            event_start = next_event_start

    @property
    def current_batch_size(self) -> int:
        return self.trainer.data_loader.global_batch_size

    def post_attach(self):
        if not self.schedule:
            return

        scheduler: Optional[Scheduler] = None
        if isinstance(self.trainer.train_module, TransformerTrainModule):
            scheduler = self.trainer.train_module.scheduler
        elif isinstance(self.trainer.train_module, TransformerPipelineTrainModule):
            scheduler = self.trainer.train_module.scheduler

        # If we have an LR scheduler, we need to make sure that the value it uses for `t_max`
        # (the end point of the schedule) won't be changed by this callback, since that would
        # lead to unexpected/discontinuous behavior in the schedule.
        # For example, if the scheduler doesn't have its own `self.t_max` set and its unit are in
        # steps, then it takes `t_max` to be `trainer.max_steps`, which will change when this callback
        # updates the batch size (unless the trainer's max duration is set in steps).
        if (
            scheduler is not None
            and not isinstance(scheduler, (ConstantScheduler, WSD))
            and getattr(scheduler, "t_max", None) is None
            and self.trainer.max_duration.unit != "steps"
            and scheduler.units != "tokens"
        ):
            raise OLMoConfigurationError(
                f"Batch size scheduler requires the {scheduler.__class__.__name__} LR scheduler's "
                "units to be set to 'tokens' unless 't_max' is set."
            )

        self._maybe_update_batch_size_and_lr()

    def post_checkpoint_loaded(self, *args):
        del args
        # NOTE: we set the "initial_lr_field" correctly after loading a checkpoint because
        # the "initial_lr_field" in the optimizer state always gets reset to the value from
        # the config, not the checkpoint.
        self._maybe_update_batch_size_and_lr()

    def pre_load_batch(self):
        self._maybe_update_batch_size_and_lr()

    def post_step(self):
        self.trainer.record_metric("train/global batch size", self.current_batch_size)

    def _maybe_update_batch_size_and_lr(self):
        if not self.enabled:
            return
        # Find latest event in the schedule to apply.
        for target_batch_size, event_start in reversed(list(zip(self.batch_sizes, self.schedule))):
            if event_start.due(
                step=self.trainer.global_step,
                tokens=self.trainer.global_train_tokens_seen,
                epoch=self.trainer.epoch,
            ):
                if target_batch_size > self.current_batch_size:
                    assert target_batch_size % self.current_batch_size == 0
                    # When increasing the batch size, need to wait until 'batches_processed' is divisible
                    # by the factor that batch size is being increased by.
                    ratio = target_batch_size // self.current_batch_size
                    if self.trainer.data_loader.batches_processed % ratio == 0:
                        self._update_batch_size_and_lr(target_batch_size)
                elif target_batch_size < self.current_batch_size:
                    assert self.current_batch_size % target_batch_size == 0
                    self._update_batch_size_and_lr(target_batch_size)

                break

    def _update_batch_size_and_lr(self, batch_size: int):
        log.info(f"Changing global batch size to {batch_size:,d}...")
        ratio = batch_size / self.current_batch_size
        lr_adjustment_factor = math.sqrt(ratio)
        self.trainer.data_loader.global_batch_size = batch_size

        optimizers: Optional[List[torch.optim.Optimizer]] = None
        scheduler: Optional[Scheduler] = None
        if isinstance(self.trainer.train_module, TransformerTrainModule):
            optimizers = [self.trainer.train_module.optim]
            scheduler = self.trainer.train_module.scheduler
        elif isinstance(self.trainer.train_module, TransformerPipelineTrainModule):
            optimizers = self.trainer.train_module.optimizers
            scheduler = self.trainer.train_module.scheduler

        if not optimizers:
            raise NotImplementedError(
                f"Unable to adjust learning rate for {self.trainer.train_module.__class__.__name__} train module class"
            )

        for optim in optimizers:
            if not isinstance(optim, (torch.optim.Adam, torch.optim.AdamW, SkipStepAdamW)):
                raise NotImplementedError(
                    f"Unable to adjust learning rate for {optim.__class__.__name__} optimizer"
                )

        log.info(
            f"Adjusting base learning rate by a factor of {lr_adjustment_factor:.4f} = sqrt({ratio})"
        )
        for optim_idx, optim in enumerate(optimizers):
            for group_idx, group in enumerate(optim.param_groups):
                lr_field = LR_FIELD if scheduler is None else scheduler.lr_field
                initial_lr_field = (
                    INITIAL_LR_FIELD if scheduler is None else scheduler.initial_lr_field
                )

                if group.get(initial_lr_field) is None:
                    group[initial_lr_field] = group[lr_field]

                group[initial_lr_field] *= lr_adjustment_factor

                # Only set the actual LR if there's no scheduler. Schedulers update the LR based
                # on the initial LR.
                if scheduler is None:
                    group[lr_field] = group[initial_lr_field]

                log.info(
                    f"Set base LR for optimizer {optim_idx+1}, group {group_idx+1} to {float(group[initial_lr_field]):.8f}"
                )

        for callback in self.trainer.callbacks.values():
            if isinstance(callback, SpeedMonitorCallback):
                log.info("Resetting speed monitor...")
                callback.reset()



from dataclasses import dataclass

from olmo_core.config import Config


@dataclass
class ExpertParallelConfig(Config):
    """
    Configuration class for expert parallelism (EP).
    """

    degree: int
    """
    The EP degree.
    """



import logging
import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

from olmo_core.config import StrEnum
from olmo_core.distributed.utils import get_rank
from olmo_core.exceptions import OLMoEnvironmentError
from olmo_core.utils import set_env_var

from .callback import Callback

if TYPE_CHECKING:
    from comet_ml import Experiment

log = logging.getLogger(__name__)

COMET_API_KEY_ENV_VAR = "COMET_API_KEY"


class CometNotificationSetting(StrEnum):
    """
    Defines the notifications settings for the Comet.ml callback.
    """

    all = "all"
    """
    Send all types notifications.
    """

    end_only = "end_only"
    """
    Only send a notification when the experiment ends (successfully or with a failure).
    """

    failure_only = "failure_only"
    """
    Only send a notification when the experiment fails.
    """

    none = "none"
    """
    Don't send any notifcations.
    """


@dataclass
class CometCallback(Callback):
    """
    Logs metrics to Comet.ml from rank 0.

    .. important::
        Requires the ``comet_ml`` package and the environment variable ``COMET_API_KEY``.

    .. note::
        This callback logs metrics from every single step to Comet.ml, regardless of the value
        of :data:`Trainer.metrics_collect_interval <olmo_core.train.Trainer.metrics_collect_interval>`.
    """

    enabled: bool = True
    """
    Set to false to disable this callback.
    """

    name: Optional[str] = None
    """
    The name to give the Comet.ml experiment.
    """

    project: Optional[str] = None
    """
    The Comet.ml project to use.
    """

    workspace: Optional[str] = None
    """
    The name of the Comet.ml workspace to use.
    """

    tags: Optional[List[str]] = None
    """
    Tags to assign the experiment.
    """

    config: Optional[Dict[str, Any]] = None
    """
    The config to save to Comet.ml.
    """

    cancel_tags: Optional[List[str]] = field(
        default_factory=lambda: ["cancel", "canceled", "cancelled"]
    )
    """
    If you add any of these tags to an experiment on Comet.ml, the run will cancel itself.
    Defaults to ``["cancel", "canceled", "cancelled"]``.
    """

    cancel_check_interval: Optional[int] = None
    """
    Check for cancel tags every this many steps. Defaults to
    :data:`olmo_core.train.Trainer.cancel_check_interval`.
    """

    notifications: CometNotificationSetting = CometNotificationSetting.none
    """
    The notification settings.
    """

    failure_tag: str = "failed"
    """
    The tag to assign to failed experiments.
    """

    auto_resume: bool = False
    """
    If ``True``, an existing experiment will be resumed from a checkpoint if the experiment name
    matches.
    """

    _exp = None
    _exp_key: Optional[str] = None
    _finalized: bool = False

    @property
    def exp(self) -> "Experiment":
        return self._exp  # type: ignore

    @exp.setter
    def exp(self, exp: "Experiment"):
        self._exp = exp

    @property
    def finalized(self) -> bool:
        return self._finalized

    def state_dict(self) -> Dict[str, Any]:
        return {"experiment_key": self._exp_key, "name": self.name}

    def load_state_dict(self, state_dict: Dict[str, Any]):
        if self.auto_resume and self.name == state_dict.get("name"):
            self._exp_key = state_dict.get("experiment_key")

    def finalize(self):
        if not self.finalized:
            self.exp.end()
            self._finalized = True

    def pre_train(self):
        if self.enabled and get_rank() == 0:
            set_env_var("COMET_DISABLE_AUTO_LOGGING", "1")

            import comet_ml as comet

            if COMET_API_KEY_ENV_VAR not in os.environ:
                raise OLMoEnvironmentError(f"missing env var '{COMET_API_KEY_ENV_VAR}'")

            if self.auto_resume and self._exp_key is not None:
                log.info(f"Resuming Comet logging from existing experiment '{self._exp_key}'")
                self.exp = cast(
                    "Experiment",
                    comet.start(
                        api_key=os.environ[COMET_API_KEY_ENV_VAR],
                        mode="get",
                        experiment_key=self._exp_key,
                        experiment_config=comet.ExperimentConfig(
                            auto_output_logging="simple",
                            display_summary_level=0,
                        ),
                    ),
                )
            else:
                self.exp = comet.Experiment(
                    api_key=os.environ[COMET_API_KEY_ENV_VAR],
                    project_name=self.project,
                    workspace=self.workspace,
                    auto_output_logging="simple",
                    display_summary_level=0,
                )
                self._exp_key = self.exp.get_key()

                if self.name is not None:
                    self.exp.set_name(self.name)

                if self.tags:
                    self.exp.add_tags(self.tags)

                if self.config is not None:
                    self.exp.log_parameters(self.config)

            if self.notifications == CometNotificationSetting.all:
                self.exp.send_notification(
                    f"Experiment {self.exp.get_name()} ({self.exp.get_key()})",
                    status="started",
                )

    def log_metrics(self, step: int, metrics: Dict[str, float]):
        if self.enabled and get_rank() == 0:
            self.exp.log_metrics(metrics, step=step)

    def post_step(self):
        cancel_check_interval = self.cancel_check_interval or self.trainer.cancel_check_interval
        if self.enabled and get_rank() == 0 and self.step % cancel_check_interval == 0:
            self.trainer.run_bookkeeping_op(
                self.check_if_canceled,
                allow_multiple=False,
                distributed=False,
            )

    def post_train(self):
        if self.enabled and get_rank() == 0:
            log.info("Finalizing successful Comet.ml experiment...")
            if self.notifications in (
                CometNotificationSetting.all,
                CometNotificationSetting.end_only,
            ):
                self.exp.send_notification(
                    f"Experiment {self.exp.get_name()} ({self.exp.get_key()})",
                    status="completed successfully",
                )
            self.finalize()

    def on_error(self, exc: BaseException):
        del exc
        if self.enabled and get_rank() == 0:
            log.warning("Finalizing failed Comet.ml experiment...")
            self.exp.add_tag(self.failure_tag)
            if self.notifications in (
                CometNotificationSetting.all,
                CometNotificationSetting.end_only,
                CometNotificationSetting.failure_only,
            ):
                self.exp.send_notification(
                    f"Experiment {self.exp.get_name()} ({self.exp.get_key()})",
                    status="failed",
                )
            self.finalize()

    def check_if_canceled(self):
        if self.enabled and not self.finalized and self.cancel_tags:
            from comet_ml.api import API

            try:
                api = API(api_key=os.environ[COMET_API_KEY_ENV_VAR])
                exp = api.get_experiment_by_key(self.exp.get_key())
                assert exp is not None
                tags = exp.get_tags()
            except Exception as exc:
                log.exception(exc)
                return

            for tag in tags or []:
                if tag.lower() in self.cancel_tags:
                    self.trainer.cancel_run("canceled from Comet.ml tag")
                    return



from typing import TypeVar

T = TypeVar("T")


def beta_feature(f: T) -> T:
    """
    Mark a class or function as a beta feature.
    """
    if f.__doc__ is None:
        f.__doc__ = ""

    f.__doc__ += """

    .. warning::
        This is a beta feature! The API is subject to change even with minor and patch releases.
        If you choose to use this feature please read the `CHANGELOG <https://github.com/allenai/OLMo-core/blob/main/CHANGELOG.md>`_
        before upgrading your version of this library.

    """

    return f



import logging
import math
import warnings
from typing import List, Optional

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from torch.distributed import DeviceMesh
from torch.distributed.fsdp import fully_shard
from torch.distributed.tensor import Placement, Shard, distribute_tensor

from olmo_core.distributed.parallel import get_device_mesh_info
from olmo_core.distributed.utils import get_local_tensor
from olmo_core.exceptions import OLMoConfigurationError
from olmo_core.utils import log_once

try:
    import grouped_gemm  # type: ignore

    gmm = grouped_gemm.ops.gmm
except ImportError:
    gmm = None

__all__ = ["MoEMLP", "DroplessMoEMLP"]


log = logging.getLogger(__name__)


class MoEMLPBase(nn.Module):
    def __init__(
        self,
        *,
        d_model: int,
        hidden_size: int,
        num_experts: int,
    ):
        super().__init__()
        self.d_model = d_model
        self.hidden_size = hidden_size
        self.num_experts = num_experts

        self.num_local_experts = num_experts
        self.hidden_sharding_degree = 1
        self.ep_mesh: Optional[DeviceMesh] = None
        self.ep_pg: Optional[dist.ProcessGroup] = None

    def apply_ep(self, ep_mesh: DeviceMesh):
        """
        Apply expert parallelism.

        :param ep_mesh: A 1D device mesh to shard experts over.
        """
        if ep_mesh.ndim != 1:
            raise RuntimeError("expert parallel mesh must be 1 dimensional")
        self._shard_experts(ep_mesh)

    def apply_tp(self, tp_mesh: DeviceMesh, float8_enabled: bool = False):
        """
        Apply expert parallelism.

        :param tp_mesh: A 1D device mesh to shard experts over.
        """
        del float8_enabled  # TODO
        if tp_mesh.ndim != 1:
            raise RuntimeError("tensor parallel mesh must be 1 dimensional")
        self._shard_experts(tp_mesh)

    def _shard_experts(self, mesh: DeviceMesh):
        num_shards = mesh.size()
        if self.num_experts % num_shards != 0:
            raise OLMoConfigurationError(
                f"'num_experts' ({self.num_experts}) must be divisible by the expert parallel shard degree ({num_shards})."
            )

        self.ep_mesh = mesh
        self.ep_pg = mesh.get_group()
        self.num_local_experts = self.num_experts // num_shards

        placements: List[Placement] = [Shard(0)]
        self.register_parameter("w1", nn.Parameter(distribute_tensor(self.w1, mesh, placements)))  # type: ignore
        self.register_parameter("w2", nn.Parameter(distribute_tensor(self.w2, mesh, placements)))  # type: ignore
        self.register_parameter("w3", nn.Parameter(distribute_tensor(self.w3, mesh, placements)))  # type: ignore

    def prepare_experts_for_fsdp(self, *, world_mesh: DeviceMesh, **kwargs):
        """
        Should be called before wrapping this module, or a parent module, with FSDP2.
        """
        # If expert/tensor parallel is not enabled then we don't need to do anything special here.
        if self.ep_mesh is None:
            return

        if self.ep_mesh.mesh_dim_names is None:
            raise RuntimeError("mesh must have named dimensions!")

        if (dim_names := world_mesh.mesh_dim_names) is None:
            raise RuntimeError("mesh must have named dimensions!")

        # If the experts are already sharded over a data parallel dimension, we need to shard them
        # over the other data parallel dimension, otherwise `fully_shard` called with the full DP
        # mesh won't handle this module correctly.
        if (ep_mesh_dim_name := self.ep_mesh.mesh_dim_names[0]).startswith("dp"):
            # Shard local experts over the adjacent DP dimension.
            dp_replicate_dim_name = dim_names[dim_names.index(ep_mesh_dim_name) - 1]
            dp_replicate_mesh = world_mesh[dp_replicate_dim_name]

            log_once(
                log, f"Sharding local experts over {get_device_mesh_info(dp_replicate_mesh)}..."
            )
            fully_shard(self, mesh=dp_replicate_mesh, **kwargs)

    def prepare_experts_for_ddp(self, *, world_mesh: DeviceMesh):
        """
        Should be called before wrapping this module, or a parent module, with FSDP2.
        """
        # TODO: do we need to do anything special here like with FSDP?
        del world_mesh
        pass


class MoEMLP(MoEMLPBase):
    """
    A basic expert MLP module with SwiGLU activation.
    """

    def __init__(
        self,
        *,
        d_model: int,
        hidden_size: int,
        num_experts: int,
        dtype: torch.dtype = torch.float32,
        init_device: str = "cpu",
    ):
        super().__init__(d_model=d_model, hidden_size=hidden_size, num_experts=num_experts)
        # NOTE: these parameters need to have a large enough first dimension (which would be num experts)
        # in order to be sharded over big world sizes with FSDP, so we flatten the first 2 dimensions.
        self.w1 = nn.Parameter(
            torch.empty(
                num_experts * d_model,
                hidden_size,
                device=init_device,
                dtype=dtype,
            ),
        )
        self.w2 = nn.Parameter(
            torch.empty(
                num_experts * hidden_size,
                d_model,
                device=init_device,
                dtype=dtype,
            ),
        )
        self.w3 = nn.Parameter(
            torch.empty(
                num_experts * d_model,
                hidden_size,
                device=init_device,
                dtype=dtype,
            ),
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(in_features), 1/sqrt(in_features)). For details, see
        # https://github.com/pytorch/pytorch/issues/57109
        for w in (self.w1, self.w2, self.w3):
            nn.init.kaiming_uniform_(w, a=math.sqrt(5))

    def extra_repr(self):
        return f"num_experts={self.num_experts}, in_features={self.d_model}, hidden_size={self.hidden_size}"

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute the expert outputs.

        :param x: The input of shape ``(num_local_experts, N, d_model)``.
        """
        og_dtype = x.dtype

        # Scale gradients and get local tensors (in case of expert parallelism).
        # shape (all): (num_local_experts, hidden_size, d_model)
        w1, w2, w3 = (
            get_local_tensor(self.w1.view(self.num_experts, self.d_model, self.hidden_size)),
            get_local_tensor(self.w2.view(self.num_experts, self.hidden_size, self.d_model)),
            get_local_tensor(self.w3.view(self.num_experts, self.d_model, self.hidden_size)),
        )

        x = x.type_as(w1)

        # Compute the MLP.
        return torch.bmm(F.silu(torch.bmm(x, w1)) * torch.bmm(x, w3), w2).to(dtype=og_dtype)


class DroplessMoEMLP(MoEMLPBase):
    """
    A dropless expert MLP module with SwiGLU activation.
    """

    def __init__(
        self,
        *,
        d_model: int,
        hidden_size: int,
        num_experts: int,
        dtype: torch.dtype = torch.float32,
        init_device: str = "cpu",
    ):
        super().__init__(d_model=d_model, hidden_size=hidden_size, num_experts=num_experts)
        # NOTE: these parameters need to have a large enough first dimension (which would be num experts)
        # in order to be sharded over big world sizes with FSDP, so we flatten the first 2 dimensions.
        self.w1 = nn.Parameter(
            torch.empty(
                num_experts * hidden_size,
                d_model,
                device=init_device,
                dtype=dtype,
            ),
        )
        self.w2 = nn.Parameter(
            torch.empty(
                num_experts * hidden_size,
                d_model,
                device=init_device,
                dtype=dtype,
            ),
        )
        self.w3 = nn.Parameter(
            torch.empty(
                num_experts * hidden_size,
                d_model,
                device=init_device,
                dtype=dtype,
            ),
        )

        self._gmm = gmm
        if self._gmm is None:
            warnings.warn(
                "Grouped GEMM not available, so the MoE will be substantially slower. "
                "Please install the 'grouped_gemm' package if possible.\n"
                "https://github.com/tgale96/grouped_gemm"
            )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(in_features), 1/sqrt(in_features)). For details, see
        # https://github.com/pytorch/pytorch/issues/57109
        for w in (self.w1, self.w2, self.w3):
            nn.init.kaiming_uniform_(w, a=math.sqrt(5))

    @torch._dynamo.disable()
    def gmm(
        self, x: torch.Tensor, w: torch.Tensor, batch_sizes: torch.Tensor, trans_b: bool = False
    ) -> torch.Tensor:
        if self._gmm is not None:
            # grouped-gemm only accepts BF16
            return self._gmm(x.to(torch.bfloat16), w.to(torch.bfloat16), batch_sizes, trans_b=trans_b)  # type: ignore
        else:
            out = []
            start = 0
            for i, size in enumerate(batch_sizes.cpu().numpy()):
                rhs = w[i, :, :].t() if trans_b else w[i, :, :]
                out.append(x[start : start + size, :] @ rhs)
                start += size
            return torch.cat(out)

    def forward(self, x: torch.Tensor, batch_size_per_expert: torch.Tensor) -> torch.Tensor:
        """
        Compute the expert outputs.

        :param x: The input of shape ``(*, d_model)``.
        :param batch_size_per_expert: Specifies how many items/tokens go to each expert. Should be a
            1-D ``LongTensor``.
        """
        # Scale gradients and get local tensors (in case of expert parallelism).
        # shape (all): (num_local_experts, hidden_size, d_model)
        w1, w2, w3 = (
            get_local_tensor(self.w1.view(self.num_experts, self.hidden_size, self.d_model)),
            get_local_tensor(self.w2.view(self.num_experts, self.hidden_size, self.d_model)),
            get_local_tensor(self.w3.view(self.num_experts, self.hidden_size, self.d_model)),
        )

        # Compute the MLP.
        x1 = self.gmm(x, w1, batch_size_per_expert, trans_b=True)
        x2 = self.gmm(x, w3, batch_size_per_expert, trans_b=True)
        x1 = F.silu(x1) * x2
        return self.gmm(x1, w2, batch_size_per_expert)
